@@ -29,7 +29,7 @@
 #include <bonobo-activation/bonobo-activation-base-service.h>
 #include <bonobo-activation/Bonobo_ActivationContext.h>
 
-#define BONOBO_ACTIVATION_FACTORY_TIMEOUT 1000
+#define BONOBO_ACTIVATION_FACTORY_PRIVATE_TIMEOUT 1000
 
 extern GStaticRecMutex _bonobo_activation_guard;
 #define BONOBO_ACTIVATION_LOCK()   g_static_rec_mutex_lock   (&_bonobo_activation_guard)
@@ -38,10 +38,10 @@ extern GStaticRecMutex _bonobo_activation_guard;
 void         bonobo_activation_timeout_reg_check_set  (gboolean           on);
 gboolean     bonobo_activation_timeout_reg_check      (gpointer           data);
 
+void         bonobo_activation_server_fork_init  (gboolean                            threaded);
 typedef CORBA_Object (*BonoboForkReCheckFn)      (const Bonobo_ActivationEnvironment *environemnt,
                                                   const char                         *act_iid,
-                                                  gpointer                            user_data,
-                                                  CORBA_Environment                  *ev);
+                                                  gpointer                            user_data);
 CORBA_Object bonobo_activation_server_by_forking (const char                        **cmd, 
 						  gboolean                            set_process_group,
 						  int                                 fd_arg,
@@ -82,5 +82,17 @@ void Bonobo_ActivationEnvValue_copy (Bonobo_ActivationEnvValue *dest,
 
 Bonobo_ActivationClient bonobo_activation_client_get (void);
 
+#ifdef G_OS_WIN32
+
+char *_bonobo_activation_win32_replace_prefix (const char *runtime_prefix,
+                                               const char *configure_time_path);
+const char *_bonobo_activation_win32_get_prefix (void);
+const char *_bonobo_activation_win32_get_server_libexecdir (void);
+const char *_bonobo_activation_win32_get_serverinfodir (void);
+const char *_bonobo_activation_win32_get_server_confdir (void);
+const char *_bonobo_activation_win32_get_localedir (void);
+
+#endif
+
 #endif /* BONOBO_ACTIVATION_PRIVATE_H */
 
